/* File: hzview.c */

static char version[] = "hzview version 2.0 (June 18, 1991)";

/*----------------------------------------------------------------------------


  	hzview 2.0: view a hanzi (GB/BIG5) file on a tty screen
	-------------------------------------------------------

	By Fung F. Lee

	Copyright (C) 1989,1991  Fung F. Lee

	All rights reserved.
	Permission to copy and distribute verbatim copies of this
	document for non-commercial purposes is granted, but changing
	it is not allowed.
	There is absolutely no warranty for this program.

	Please send your comments and suggestions to the author:
	Fung F. Lee	lee@milo.stanford.edu

	The font files should all be under the directory HZDIR
	(setenv HZDIR <hanzi-directory>).


----------------------------------------------------------------------------*/
  

#include <stdio.h>
#include <ctype.h>
#include <string.h>

#define True 1
#define False 0

#define GBcode 0
#define BIGcode 1

#define GBspace		0x2121
#define BIGspace	0xA140

#define MAX(a, b)	((a)>=(b)? (a) : (b))
#define DB(hi,lo)	(((hi)&0xFF) << 8 | (lo)&0xFF)
#define HGB2GB(hi,lo)	(((hi)&0x7F) << 8 | (lo)&0x7F)
/* test ranges of first and second bytes */
#define isGB1(c)	((c)>=0xA1 && (c)<=0xFE)
#define isGB2(c)	((c)>=0xA1 && (c)<=0xFE)
#define isBIG1(c)	((c)>=0xA1 && (c)<=0xF9)
#define isBIG2(c)	((c)>=0x40 && (c)<=0x7E || (c)>=0xA1 && (c)<=0xFE)


extern char *GBbitmap(), *BIGbitmap();
extern void loadGBfont(), loadBIGfont();
extern char *getenv();

int encoding = GBcode;
float leftmargin = 0,   /* column 0 */  rightmargin = 78; /* column 78 */
int HZbm = 16;
int hzbyte;	/* number of bytes per hanzi image */
float HZpoint = 16, HZspacing = 2, HZskip = 18;
float Eskip = 1;
float hpos;
static char HZDIR[] = "HZDIR";
static char nullName[] = "";
char progname[] = "hzview";
char hzdir[100], fontdir[100];
char hzfont[30];
char hzName[30]="Beijing";
char *inputName;
FILE *fin;

#define MAXROW 24	/* maximum bitmap size per hanzi */
#define MAXCOL 132	/* maximum number of columns on terminal */
char screen[MAXROW][MAXCOL];
char cmark = 'D';
int hFill = False;	/* flag if hanzi screen buffer filled (not empty) */
int HZmode = False;	/* current character mode */

void usage();
void GBscanner(), BIGscanner(), fputNewLine();
void fputHanzi(), fputWord(), fputAscii();
void ttyInit(), ttyShow();


void usage()
{
    FILE *f = stderr;

    fprintf(f, "%s\n\n", version);
    fprintf(f, "usage:\t%s ", progname);
    fprintf(f, "[-gb] [-big] \\\n");
    fprintf(f, "\t[-margin <left> <right>] \\\n" );
    fprintf(f, "\t[-hf <hanzi-font> <point-size> <inter-hanzi-spacing>] \\\n");
    fprintf(f, "\t[-mark <character>] \\\n");
    fprintf(f, "\t[file]\n");
    exit(1);
}


#define LEN 100
#define PtPerIn 1

void main(argc, argv)
int argc;
char *argv[];
{
    char s[LEN];
    int i;
    FILE *header;
    float ftmp;
    
    if (getenv(HZDIR) == NULL)
    {
	fprintf(stderr, "Please setenv %s appropriately.\n", HZDIR);
	exit(1);
    }
    strcpy(hzdir, getenv(HZDIR));
    strcat(hzdir, "/");
    strcpy(fontdir, hzdir);
    inputName = nullName;
    
    for (i=1; i<argc; i++)
    {
	if (strcmp("-gb", argv[i]) == 0)
	    encoding = GBcode;
	else if (strcmp("-big", argv[i]) == 0)
	    encoding = BIGcode;
	else if (strcmp("-margin", argv[i]) == 0)
	{
	    if (++i >= argc) usage();
	    if (sscanf(argv[i], "%f", &ftmp) != 1) usage();
	    leftmargin = ftmp * PtPerIn;
	    if (++i >= argc) usage();
	    if (sscanf(argv[i], "%f", &ftmp) != 1) usage();
	    rightmargin = ftmp * PtPerIn;
	}
	else if (strcmp("-hf", argv[i]) == 0)
	{
	    if (++i >= argc) usage();
	    strcpy(hzName, argv[i]);
	    if (++i >= argc) usage();
	    if (sscanf(argv[i], "%f", &HZpoint) != 1) usage();
	    HZbm = HZpoint;
	    if (++i >= argc) usage();
	    if (sscanf(argv[i], "%f", &HZspacing) != 1) usage();
	}
	else if (strcmp("-mark", argv[i]) == 0)
	{
	    if (++i >= argc) usage();
	    if (sscanf(argv[i], "%c", &cmark) != 1) usage();
	}
	else if (argv[i][0] == '-')
	{
	    usage();
	}
	else
	    inputName = argv[i];
    }
    
    HZskip = HZpoint + HZspacing;
    
    if (leftmargin + HZskip > rightmargin)
    {
	fprintf(stderr, "Insufficient space between left and right margins.\n");
	exit(1);
    }

    sprintf(s, "%d", HZbm);
    strcpy(hzfont, hzName);
    strcat(hzfont, s);
    
    hzbyte = (HZbm * HZbm) / 8;
    if (encoding == GBcode)
	loadGBfont();
    else if (encoding == BIGcode)
	loadBIGfont();
    
    if (strcmp(inputName, nullName) == 0)
	fin = stdin;
    else if ((fin = fopen(inputName, "r")) == 0)
    {
	fprintf(stderr, "Cannot open input file %s\n", inputName);
	exit(1);
    }
    ttyInit();
    hpos = leftmargin;
    if (encoding == GBcode)
	GBscanner(fin, stdout);
    else if (encoding == BIGcode)
	BIGscanner(fin, stdout);
}


#define SLEN 80		/* maximum English string length */

void GBscanner(fin, fout)
FILE *fin, *fout;
{
    int c1, c2, i;
    unsigned int code;
    char *image, s[SLEN];
    
    fputNewLine(fout);
    while ((c1=fgetc(fin)) != EOF)
    {
	if (isGB1(c1))
	{
	    c2 = fgetc(fin);
	    code = HGB2GB(c1, c2);
	    if (!isGB2(c2))
	    {
		fprintf(stderr, "unrecognized code: 0x%2x%2x\n", c1, c2);
		code = GBspace;	/* default error recovery */
	    }
	    image = GBbitmap(code);
	    fputHanzi(image, fout);
	}
	else if (c1 & 0x80)
	    fprintf(stderr, "unrecognized character code: 0x%2x\n", c1);
	else if (isalnum(c1))
	{
	    s[0] = c1;
	    i = 1;
	    while (i<SLEN-1 && (c1=fgetc(fin)) && isalnum(c1))
		s[i++] = c1;
	    s[i++] = '\0';
	    if (i<SLEN)
		ungetc(c1, fin);
	    fputWord(s, fout);
	}
	else
	    fputAscii(c1, fout);
    }
}


void BIGscanner(fin, fout)
FILE *fin, *fout;
{
    int c1, c2, i;
    unsigned int code;
    char *image, s[SLEN];
    
    fputNewLine(fout);
    while ((c1=fgetc(fin)) != EOF)
    {
	if (isBIG1(c1))
	{
	    c2 = fgetc(fin);
	    code = DB(c1, c2);
	    if (!isBIG2(c2))
	    {
		fprintf(stderr, "unrecognized code: 0x%2x%2x\n", c1, c2);
		code = BIGspace;	/* default error recovery */
	    }
	    image = BIGbitmap(code);
	    fputHanzi(image, fout);
	}
	else if (c1 & 0x80)
	    fprintf(stderr, "unrecognized character code: 0x%2x\n", c1);
	else if (isalnum(c1))
	{
	    s[0] = c1;
	    i = 1;
	    while (i<SLEN-1 && (c1=fgetc(fin)) && isalnum(c1))
		s[i++] = c1;
	    s[i++] = '\0';
	    if (i<SLEN)
		ungetc(c1, fin);
	    fputWord(s, fout);
	}
	else
	    fputAscii(c1, fout);
    }
}


void fputNewLine(f)
FILE *f;
{
    int i;

    putc('\n', f);
    if (hFill)
       ttyShow(f);
    hFill = False;
    for (i=0; i<leftmargin; i++)
	putc(' ', f);
    hpos = leftmargin;

}


void fputHanzi(image, f)
char *image;	/* image points to hzbyte bytes of bitmap */
FILE *f;
{
    int row, col, k=0;
    char w;

    if (!HZmode)
	fputNewLine(f);
    HZmode = True;

    if (hpos + HZskip > rightmargin)
	fputNewLine(f);
    for (row = 0; row < HZbm; row++)
    {
	for (col = hpos; col < hpos + HZbm; col++)
	{
	    if (k==0) {w = *image++; k = 8;}
	    screen[row][col] = (w & 0x80) ? cmark : ' ';
	    w <<= 1; k--;
	}
    }
    hFill = True;
    hpos += HZskip;
}


void fputWord(s, f)
char *s;
FILE *f;
{
    if (HZmode)
	fputNewLine(f);
    HZmode = False;

    if (hpos + strlen(s) * Eskip > rightmargin)
	fputNewLine(f);
    fprintf(f, "%s", s);
    hpos += strlen(s) * Eskip;
}


void fputAscii(c, f)
char c;
FILE *f;
{
    if (HZmode)
	fputNewLine(f);
    HZmode = False;

    if (c=='\n')
    {
	fputNewLine(f);
	return;
    }
    if (c=='\t')
    {
	c = ' ';
    }
    if (hpos + Eskip > rightmargin)
	fputNewLine(f);
    fputc(c, f);
    hpos += Eskip; 
}


void ttyInit()
{
  int row, col, k;

  for (row = 0; row < MAXROW; row++)
    for (col = 0; col < MAXCOL; col++)
      screen[row][col] = ' ';
}


void ttyShow(f)
FILE *f;
{
  int row, col, k;

  if (hpos == leftmargin) return;
  for (row = 0; row < HZbm; row++)
    {
      for (col = 0; col < hpos; col++)
	putc(screen[row][col], f);
      putc('\n', f);
    }	
}




